/**
 * CRFalign: A CRF-based sentence aligner for parallel corpora.
 * 
 * Copyright (C) 2012  Eva Mujdricza-Maydt
 * Copyright (C) 2012  Huiqin Körkel-Qu
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Developers:
 * - Eva Mujdricza-Maydt (mujdricz@cl.uni-heidelberg.de)
 * - Huiqin Körkel-Qu (huiqin@cl.uni-heidelberg.de)
 */
 
package corpora;

import com.cybozu.labs.langdetect.LangDetectException;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.regex.Matcher;
import javax.xml.parsers.ParserConfigurationException;
import objects.Alignment_DocLevel;
import objects.Alignment_SentLevel;
import objects.Sentence;
import org.xml.sax.SAXException;
import java.util.Comparator;

/** Extraction methods for the evaluation part.
 *
 * @author Eva Mujdricza-Maydt
 * @lastChange 20120626
 */
public class ExtractCorpus_forEvaluation {

	
	/** Write alignments for one file-pair in opus format.
	 * E.g.
	 * <br/><br/>
	 * 1;1<br/>
	 * 2 3;2 3 4<br/>
	 * 4;<br/>
	 * 5;5<br/>
	 * <br/><br/>
	 * NOTE that if there is no number on one of the sides, this means zero alignment
	 * 
	 * 
	 * @param outputFileName name of the output file containing the alignments
	 * @param opusAlignment	alignments for one file-pair
	 * @throws FileNotFoundException
	 * @throws IOException 
	 */
	public static void writeAlignmentFile_opusFormat(
		String outputFileName, Alignment_DocLevel opusAlignment)
		throws FileNotFoundException, IOException {

		utils.Helper.checkFileForWriting(outputFileName);
		FileWriter fw = new FileWriter(new File(outputFileName));

		ArrayList<Alignment_SentLevel> sentAlignments = opusAlignment.getSentAlignments();

		for (Alignment_SentLevel sa : sentAlignments) {
			fw.write(sa.getSourceSentAlign_asString());
			fw.write(corpora.opensubtitles.Constants_OpenSubtitles.SEPARATOR_opus_link);
			fw.write(sa.getTargetSentAlign_asString());
			fw.write("\n");
			fw.flush();
		}

		fw.close();
	}
	
	
	/** Gutenberg data entails only partial alignments.
	 * <br/>NOTE that the original Gargantua annotation entails only "partial
	 * alignments" i.e. one-to-one alignments also in case of many-to-many
	 * alignments. In this case more one-to-one-alignments should be merged to 
	 * one many-to-many alignment. The Gargantua annotation entails 
	 * no zero alignments at all.
	 * 
	 * @param alignmentFileName file name of the Gargantua alignment file with path
	 * @return list of all partial alignments read in the alignment file
	 * @throws FileNotFoundException
	 * @throws IOException 
	 */
	public static ArrayList<int[]> readInPartialAlignments_gargantuaAnnotation(
		String inputFileName)
		throws FileNotFoundException, IOException {
	
		BufferedReader br = new BufferedReader(new FileReader(new File(inputFileName)));

		//collect the numbers
		String line = null;
		ArrayList<int[]> listWith1To1Alignments = new ArrayList<int[]>();
		while ((line = br.readLine()) != null) {
			//<s>de:2 en:1</s>
			if (line.startsWith("<s>")) {
				Matcher m = corpora.gutenberg.Constants_Gutenberg.PATTERN_ALIGNMENTS.matcher(line);
				if (m.matches()) {

					int[] l = new int[2];
					l[0] = Integer.parseInt(m.group(1));
					l[1] = Integer.parseInt(m.group(2));
					listWith1To1Alignments.add(l);

				} else {
					String msg = "\nNo match of line: " + line + "\n";
					throw new NullPointerException(msg);
				}
			}
		}
		br.close();

		return listWith1To1Alignments;
	}


	/** Reads partial alignments from the hunalign output.
	 * <br/>
	 * NOTE that hunalign also aligns empty sentences, e.g. if there is an 
	 * empty line at the end of the documents, it will be considered.
	 * <br/>NOTE that hunalign has a strange ladder format
	 * 
	 * Original ladder interpretation can be read in the hunalign software: 
	 * hunalign-1.1/src/hunalign/alignerTool.cpp
	 * (rows 255-260):
	<pre>  {@literal 
	for ( int i=0; i<bestTrail.size()-1; ++i )
	{
	// The [huPos, nexthuPos) interval corresponds to the [enPos, nextenPos) interval.
	int huPos = bestTrail[i].first;
	int enPos = bestTrail[i].second;
	int nexthuPos = bestTrail[i+1].first;
	int nextenPos = bestTrail[i+1].second;
	
	if (textual)
	{
	int j;
	for ( j=huPos; j<nexthuPos; ++j )
	{
	os << huSentenceListPretty[j].words;
	
	if (j+1<nexthuPos)
	os << " ~~~ ";
	}
	
	os << "\t" ;
	
	for ( j=enPos; j<nextenPos; ++j )
	{
	os << enSentenceList[j].words;
	if (j+1<nextenPos)
	{
	os << " ~~~ ";
	}
	}
	}
	else // (!textual)
	{
	os << huPos << "\t" << enPos ;
	}
	
	os << "\t" << trailScoresInterval(i);
	
	os << std::endl;
	} }</pre>
	 * 
	 * @param currInputFileName the hunalign alignment output file name with path
	 */
	public static ArrayList<int[]> readInPartialAlignments_hunalignAnnotation(
		String currInputFileName)
		throws FileNotFoundException, IOException {

		BufferedReader br = new BufferedReader(new FileReader(new File(currInputFileName)));

		//collect the numbers
		String line = null;
		ArrayList<int[]> bestTrails = new ArrayList<int[]>();
		ArrayList<int[]> listWithPartialAlignments = new ArrayList<int[]>();
		while ((line = br.readLine()) != null) {

			if (line.isEmpty()) {
				continue;
			}

			//12 \t 13 \t 0.87665
			String[] splittedLine = line.trim().split(annotators.hunalign.Constants_Hunalign.SEPARATOR_HUNALIGN_ALIGNMENT);

			if (splittedLine.length != 3) {
				String msg = "\nStrange hunalign line: \"" + line + "\"\n";
				throw new ArrayIndexOutOfBoundsException(msg);
			}

			int[] l = new int[2];
			l[0] = Integer.parseInt(splittedLine[0]) + 1; //!!! hunalign numbers start with 0
			l[1] = Integer.parseInt(splittedLine[1]) + 1;
			bestTrails.add(l);
		}
		br.close();
		
		//add a dummy number pair to the bestTrails as last entry
		int[] dummy = new int[2];
		dummy[0] = bestTrails.get(bestTrails.size()-1)[0]+1;
		dummy[1] = bestTrails.get(bestTrails.size()-1)[1]+1;
		bestTrails.add(dummy);
		
		for (int i = 0; i < bestTrails.size() - 1; ++i) {
			// The [huPos, nexthuPos) interval corresponds to the [enPos, nextenPos) interval.
			int srcPos = bestTrails.get(i)[0];
			int trgPos = bestTrails.get(i)[1];
			int nextSrcPos = bestTrails.get(i + 1)[0];
			int nextTrgPos = bestTrails.get(i + 1)[1];

			if (srcPos < nextSrcPos) {
				for (int j = srcPos; j < nextSrcPos; ++j) {
					if (trgPos < nextTrgPos) {
						for (int k = trgPos; k < nextTrgPos; ++k) {
							int[] st = new int[]{j, k};
							listWithPartialAlignments.add(st);
						}
					} else {
						int[] st = new int[2];
						st[0] = j;
						st[1] = main.Constants_Main.MINIMUMSENTNUMBER-1;
						listWithPartialAlignments.add(st);
					}
				}
			} else {
				for (int k = trgPos; k < nextTrgPos; ++k) {
					int[] st = new int[2];
					st[0] = main.Constants_Main.MINIMUMSENTNUMBER-1;
					st[1] = k;
					listWithPartialAlignments.add(st);
				}	
			}
		}

		//!!!
		//remove the last one, because the documents all (???) ends with 
		//an empty line
		int last_src = listWithPartialAlignments.get(listWithPartialAlignments.size() - 1)[0];
		int last_trg = listWithPartialAlignments.get(listWithPartialAlignments.size() - 1)[1];
		ArrayList<int[]> listWithPartialAlignments_good = new ArrayList<int[]>();
		for (int i = 0; i < listWithPartialAlignments.size(); i++) {
			int[] st = listWithPartialAlignments.get(i);
			if (st[0] == last_src || st[1] == last_trg) {
				//nothing
			} else {
				listWithPartialAlignments_good.add(st);
			}
		}
		
		for (int[] st : listWithPartialAlignments_good) {
			if ((st[0] < main.Constants_Main.MINIMUMSENTNUMBER) && st[1] < (main.Constants_Main.MINIMUMSENTNUMBER)) {
				String msg = "\nZero-zero alignment in " + currInputFileName + "!\n";
				throw new NullPointerException(msg);
			}
		}
		
		listWithPartialAlignments_good = sortPartialAlignmentList(listWithPartialAlignments_good);
		return listWithPartialAlignments_good;
	}


	/** Sorts the list of partial alignments.
	 * 
	 * @param partialAl partial alignment list
	 * @return the sorted partial alignment list
	 */
	private static ArrayList<int[]> sortPartialAlignmentList(
		ArrayList<int[]> partialAl) {

		Collections.sort(partialAl, COMPARATOR_PARTIALALIGNMENT);
		return partialAl;
	}

	/**
	 * Comparator for partial alignments. The alignments are in OPUS-style
	 * format (see README.txt).
	 */
	private static class PartialAlignmentComparator implements Comparator<int[]> {

		private PartialAlignmentComparator() {
		}

		@Override
		public int compare(int[] o1, int[] o2) {
			int cmp;
			
			int o1src = o1[0];
			int o1trg = o1[1];
			int o2src = o2[0];
			int o2trg = o2[1];
			
			//_;? to ?;?
			if (o1src < main.Constants_Main.MINIMUMSENTNUMBER) {
				
				if (o1trg < main.Constants_Main.MINIMUMSENTNUMBER) {
					String msg = "\nPartial sentence alignment entails unallowed zero-zero-alignment: " + o1src + "," + o1trg + "\n";
					throw new NullPointerException(msg);
				}
				
				//_;x1 _;?
				//only target sides are to compare
				if (o2src < main.Constants_Main.MINIMUMSENTNUMBER) {
					
					if (o2trg < main.Constants_Main.MINIMUMSENTNUMBER) {
						String msg = "\nPartial sentence alignment entails unallowed zero-zero-alignment: " + o2src + "," + o2trg + "\n";
						throw new NullPointerException(msg);
					}
					//_;x1 _;y2
					cmp = getComparisonBtwFirstInts(o1trg, 1, o2trg, 1);
				}
				//_;x1 y1;?
				else {
					//_;x1 y1;_
					//nothing to compare
					if (o2trg < main.Constants_Main.MINIMUMSENTNUMBER) {
						cmp = 0;
					} 
					//_;x1 y1;y2
					//only target sides are to compare
					else {
						cmp = getComparisonBtwFirstInts(o1trg, 1, o2trg, 1);
					}
				}
			} 
			//x1;? to ?;?
			else {
				//x1;_ to ?;?
				if (o1trg < main.Constants_Main.MINIMUMSENTNUMBER) {
					//x1;_ to _;?
					if (o2src < main.Constants_Main.MINIMUMSENTNUMBER) {
						
						if (o2trg < main.Constants_Main.MINIMUMSENTNUMBER) {
							String msg = "\nPartial sentence alignment entails unallowed zero-zero-alignment: " + o2trg + "," + o2trg + "\n";
							throw new NullPointerException(msg);
						}
						
						//x1;_ to _;y2
						cmp = 0;
					}
					//x1;_ to y1;?
					else {
						//only source sides are to compare
						cmp = getComparisonBtwFirstInts(o1src, 1, o2src, 1);
					}
				} 
				//x1;x2 to ?;?
				else {
					//x1;x2 to _;?
					if (o2src < main.Constants_Main.MINIMUMSENTNUMBER) {
						
						if (o2trg < main.Constants_Main.MINIMUMSENTNUMBER) {
							String msg = "\nPartial sentence alignment entails unallowed zero-zero-alignment: " + o2trg + "," + o2trg + "\n";
							throw new NullPointerException(msg);
						}
						//x1;x2 to _;y2
						//only target sides are to compare
						cmp = getComparisonBtwFirstInts(o1trg, 1, o2trg, 1);
					}
					//x1;x2 to y1;?
					else {
						//x1;x2 to y1;_
						//only source sides are to compare
						if (o2trg < main.Constants_Main.MINIMUMSENTNUMBER) {
							cmp = getComparisonBtwFirstInts(o1src, 1, o2src, 1);
						}
						//x1;x2 to y1;y2
						else {
							if (o1src < o2src) {
								cmp = -1;
							} else if (o1src > o2src) {
								cmp = 1;
							} else {
								cmp = getComparisonBtwFirstInts(o1trg, 1, o2trg, 1);
							}
						}
					}
				}
			}

			return cmp;
		}
	}
	
	/** Comparator instance for sorting partial alignments. */
	private final static Comparator<int[]> COMPARATOR_PARTIALALIGNMENT = 
		new PartialAlignmentComparator();


	/** Sorts an sentence-level alignment list. 
	 * 
	 * @param al list of sentence-level alignments
	 * @return the sorted sentence-level alignment list
	 */
	public static ArrayList<Alignment_SentLevel> sortAlignmentList(
		ArrayList<Alignment_SentLevel> al) {

		try {
			Collections.sort(al, COMPARATOR_SENTALIGNMENT);
		} catch (IllegalArgumentException iae) {
			System.out.println("ILLEGALARGUMENTEXCEPTION");
		}
		return al;
	}

	/**
	 * Comparator for sentence-level alignments.
	 */
	private static class SentAlignmentComparator implements Comparator<Alignment_SentLevel> {

		private SentAlignmentComparator() {
		}

		@Override
		public int compare(Alignment_SentLevel o1, Alignment_SentLevel o2) {
			if (o1.hasSourceSentAlign() == false || o1.hasTargetSentAlign() == false) {
				throw new NullPointerException("Not set first alignment list to compare!");
			}
			if (o2.hasSourceSentAlign() == false || o2.hasTargetSentAlign() == false) {
				throw new NullPointerException("Not set second alignment list to compare!");
			}
			int cmp = 2;

			ArrayList<Integer> o1src = o1.getSourceSentAlign();
			ArrayList<Integer> o1trg = o1.getTargetSentAlign();
			ArrayList<Integer> o2src = o2.getSourceSentAlign();
			ArrayList<Integer> o2trg = o2.getTargetSentAlign();

			//note that we assume, that the numbers are sorted within the target side
			//and so the number with index zero is the lowest value
			//_;? to ?;?
			if (o1.hasEmptySourceSentAlign()) {

				if (o1.hasEmptyTargetSentAlign()) {
					String msg = "\nNo possible alignment: both source and target side empty.\n";
					throw new NullPointerException(msg);
				}

				//_;x2 to ?;y2
				//only target aligns to compare
				//independent of the source side of the o2
				if (!(o2.hasEmptyTargetSentAlign())) {
					cmp = getComparisonBtwFirstInts(o1trg.get(0), o1trg.size(), o2trg.get(0), o2trg.size());
				} 
				//_;x2 to ?;_
				else {
					//e.g. _;12 and 15;_ to compare
					//--> nothing to compare
					//???
					cmp = 0;
				}
			} 
			//o1src is not empty
			//x1;? to ?;?
			else {
				//x1;_ to ?;?
				if (o1.hasEmptyTargetSentAlign()) {
					//x1;_ to _;?
					if (o2.hasEmptySourceSentAlign()) {
						//e.g. 15;_ and _;12 to compare
						//--> nothing to compare
						//???
						cmp = 0;
						
					//NOTE that it is no matter if o2.hasEmptyTargetSentAlign() == true
					} 
					//x1;_ to y2;?
					else {
						cmp = getComparisonBtwFirstInts(o1src.get(0), o1src.size(), o2src.get(0), o2src.size());
					}
				} 
				//both sides of o1 are filled
				//x1;x2 to ?;?
				else {
					//x1;x2 to _;y2
					//only compare the target sides
					if (o2.hasEmptySourceSentAlign()) {

						if (o2.hasEmptyTargetSentAlign()) {
							String msg = "\nNo possible alignment: both source and target side empty.\n";
							throw new NullPointerException(msg);
						}
						cmp = getComparisonBtwFirstInts(o1trg.get(0), o1trg.size(), o2trg.get(0), o2trg.size());
					} 
					//x1;x2 to y1;?
					else {
						//x1;x2 to y1;_
						//only compare the source sides
						if (o2.hasEmptyTargetSentAlign()) {
							cmp = getComparisonBtwFirstInts(o1src.get(0), o1src.size(), o2src.get(0), o2src.size());
						} 
						//x1;x2 to y1;y2
						else {
							if (o1src.get(0) < o2src.get(0)) {
								cmp = -1;
							} else if (o1src.get(0) > o2src.get(0)) {
								cmp = 1;
							} 
							else {
								cmp = getComparisonBtwFirstInts(o1trg.get(0), o1trg.size(), o2trg.get(0), o2trg.size());
							}
						}
					}
				}
			}

			
			if (cmp == 2) {
				StringBuilder sb1 = new StringBuilder();
				sb1.append(o1.getSourceSentAlign_asString()).append(":").append(o1.getTargetSentAlign_asString());
				StringBuilder sb2 = new StringBuilder();
				sb2.append(o2.getSourceSentAlign_asString()).append(":").append(o2.getTargetSentAlign_asString());
				String msg = "\nNo comparison available for o1: " + sb1.toString() + " and o2: " + sb2.toString() + "\n";
				throw new NullPointerException(msg);
			}

			return cmp;
		}
	}
	
	/** Comparator instance for sentence-level alignments. */
	private final static Comparator<Alignment_SentLevel> COMPARATOR_SENTALIGNMENT = 
		new SentAlignmentComparator();

	/** Auxiliary method for the alignment comparators.
	 * 
	 * @param nro1 the first integer of the first sentence number list
	 * @param o1Size size of the first sentence number list
	 * @param nro2 the first integer of the second sentence number list
	 * @param o2Size size of the second sentence number list
	 * @return -1, if the first list is earlier in the order, 
	 *		+1, if the second list is to be placed earlier, 
	 *		0 if the two lists are equal
	 */
	private static int getComparisonBtwFirstInts(
		int nro1, int o1Size, int nro2, int o2Size) {
		int cmp = 2;
		
		if (nro1 < nro2) {
			cmp = -1;
		} else if (nro1 > nro2) {
			cmp = 1;
		} else {
			//longer is later
			if (o1Size < o2Size) {
				cmp = -1;
			} else if (o1Size > o2Size) {
				cmp = 1;
			} else {
				cmp = 0;
			}
		}
		
		return cmp;
	}
	
	/** Writes a html-file for each text pair according to the given alignment file.
	 * NOTE that if not all sentences of the texts are aligned in the 
	 * given alignment file, then the missing sentences will be completed as
	 * zero alignments according to the TreeTagger output files.
	 * 
	 * @param dataSetPath path of the current data set
	 * @param alignmentExtension extension of the alignment files
	 * @param treetaggeroutputExtension extension of the TreeTagger xml-output
	 * @param outputPath path for the output html-files
	 * @throws FileNotFoundException
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws LangDetectException 
	 */
	public static void writeParallelHtmlFiles(
		String dataSetPath,
		String alignmentExtension,
		String treetaggeroutputExtension,
		String outputPath)
		throws FileNotFoundException, IOException,
		ParserConfigurationException, SAXException, LangDetectException {

		String alignmentPath = dataSetPath + main.Constants_Main.FOLDER_ALIGNMENTS;
		ArrayList<String> fileNameStems = utils.Helper.extractFileNamesByEnding_woEnding(alignmentPath, alignmentExtension);
		
		for (String fileStem : utils.Helper.sortAL_str(fileNameStems)) {

			String givenAnnotation = alignmentPath + fileStem + alignmentExtension;

			try {
				utils.Helper.checkFileForReading(givenAnnotation);
			} catch (FileNotFoundException fne) {
				System.out.println("\nNO ANNOTATION FILE FOR " + fileStem + " WITH EXTENSION " + alignmentExtension);
				continue;
			}

			Alignment_DocLevel opusDocAl = eval.EvaluateCorpus.readInAlignment(
				alignmentExtension, givenAnnotation, fileStem, fileStem);

			ArrayList<Sentence> srcSentences = annotators.treetagger.ExtractCorpus_forTreeTagger.readInTreeTaggerOutput_xml(
				dataSetPath + main.Constants_Main.FOLDER_SRC + fileStem + treetaggeroutputExtension);

			ArrayList<Sentence> trgSentences = annotators.treetagger.ExtractCorpus_forTreeTagger.readInTreeTaggerOutput_xml(
				dataSetPath + main.Constants_Main.FOLDER_TRG + fileStem + treetaggeroutputExtension);

			//complete the original alignments with zero-alignments
			//if there is no alignment for a sentence
			opusDocAl.setSentAlignments(completeAlignmentsWithZeroAligments(opusDocAl.getSentAlignments(), srcSentences, trgSentences));
			
			
			String output_html = outputPath + fileStem + alignmentExtension + main.Constants_Main.EXT_HTML;
			utils.Helper.checkFileForWriting(output_html);
			FileWriter fw = new FileWriter(new File(output_html));

			fw.write("<html>\n<head><title>" + fileStem + "_" + fileStem + "</title></head>\n<body>\n");

			for (Alignment_SentLevel oneAl : opusDocAl.getSentAlignments()) {

				ArrayList<Integer> currSrcNrs = oneAl.getSourceSentAlign();

				for (Integer i : currSrcNrs) {
					String si = srcSentences.get(i - 1).getWordListAsString_suppressException();
					StringBuilder sb = new StringBuilder();
					sb.append("src (").append(i).append("):\t").append(si).append("<br/>\n");
					fw.write(sb.toString());
					fw.flush();
				}

				ArrayList<Integer> currTrgNrs = oneAl.getTargetSentAlign();
				for (Integer i : currTrgNrs) {
					String si = trgSentences.get(i - 1).getWordListAsString_suppressException();
					StringBuilder sb = new StringBuilder();
					sb.append("trg (").append(i).append("):\t").append(si).append("<br/>\n");
					fw.write(sb.toString());
					fw.flush();
				}
				fw.write("<hr/>\n");
			}
			fw.write("</body></html>");
			fw.flush();
			fw.close();
			System.out.println("-> output written: " + output_html);
		}
	}
	
	/** Completes the list of sentence-level alignments with zero-alignments, if
	 * there are any.
	 * 
	 * @param originalAlignments list of sentence-level alignments
	 * @param srcSentences all sentences in the source file
	 * @param trgSentences all sentences in the target file
	 * @return sorted list of sentence-level alignments extended by zero-alignments, if 
	 *			any not existing in the input alignment list
	 */
	private static ArrayList<Alignment_SentLevel> completeAlignmentsWithZeroAligments(
		ArrayList<Alignment_SentLevel> originalAlignments,
		ArrayList<Sentence> srcSentences,
		ArrayList<Sentence> trgSentences){
	
		int prevSrcNr = 0;
		int prevTrgNr = 0;
		ArrayList<Alignment_SentLevel> missingZeros = new ArrayList<Alignment_SentLevel>(0);
		
		for (Alignment_SentLevel alignment : originalAlignments) {
			ArrayList<Integer> currSrcAl = alignment.getSourceSentAlign();
			ArrayList<Integer> currTrgAl = alignment.getSourceSentAlign();
			
			if (currSrcAl.get(0) < prevSrcNr+1) {
				for (int i = prevSrcNr+1; i < currSrcAl.get(0); i++) {
					Alignment_SentLevel newAl = Alignment_SentLevel.newInstance();
					newAl.setSourceSentAlign(i);
					newAl.setEmptyTargetSentAlign();
					missingZeros.add(newAl);
				}
			}
			prevSrcNr = currSrcAl.get(currSrcAl.size()-1);
			
			if (currTrgAl.get(0) < prevTrgNr+1) {
				for (int i = prevTrgNr+1; i < currTrgAl.get(0); i++) {
					Alignment_SentLevel newAl = Alignment_SentLevel.newInstance();
					newAl.setEmptySourceSentAlign();
					newAl.setTargetSentAlign(i);
					missingZeros.add(newAl);
				}
			}
			prevTrgNr = currTrgAl.get(currTrgAl.size()-1);
		}
		
		//probably there are missing alignments at the end of the texts
		for (int i = prevSrcNr+1; i <= srcSentences.size(); i++) {
			Alignment_SentLevel newAl = Alignment_SentLevel.newInstance();
			newAl.setSourceSentAlign(i);
			newAl.setEmptyTargetSentAlign();
			missingZeros.add(newAl);
		}
		for (int i = prevTrgNr+1; i <= trgSentences.size(); i++) {
			Alignment_SentLevel newAl = Alignment_SentLevel.newInstance();
			newAl.setEmptySourceSentAlign();
			newAl.setTargetSentAlign(i);
			missingZeros.add(newAl);
		}
			
		if (!missingZeros.isEmpty()) {
			originalAlignments.addAll(missingZeros);
			originalAlignments = sortAlignmentList(originalAlignments);
		}
		return originalAlignments;
	}
}
