/**
 * CRFalign: A CRF-based sentence aligner for parallel corpora.
 * 
 * Copyright (C) 2012  Eva Mujdricza-Maydt
 * Copyright (C) 2012  Huiqin Körkel-Qu
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Developers:
 * - Eva Mujdricza-Maydt (mujdricz@cl.uni-heidelberg.de)
 * - Huiqin Körkel-Qu (huiqin@cl.uni-heidelberg.de)
 */
 
package corpora.opensubtitles;

import java.util.regex.Pattern;

/** Often used variables and constants for OpenSubtitles corpus.
 * <br/>NOTE that "OpenSubtitles" and "Opus" corpus mean the same.
 *
 * @author Eva Mujdricza-Maydt
 * @lastChange 20120626
 */
public class Constants_OpenSubtitles {
	
	
	/** xml-element incorporating all alignments for one parallel movie-pair
	 * <br/>example for a link-group:
	 * {@literal <linkGrp targType="s" 
	 toDoc="en/Action/1968/2565_150326_205118_cera_una_volta_il_west.xml.gz" 
	 fromDoc="de/Action/1968/2565_132358_181947_cera_una_volta_il_west.xml.gz" />}
	 */
	static final String ELEMENT_linkGrp = "linkGrp";
	/** xml-attribute for zipped file name (path) of current movie title in source language */
	static final String ATTRIBUTE_fromDoc = "fromDoc";
	/** xml-attribute for zipped file name (path) of current movie title in target language */
	static final String ATTRIBUTE_toDoc = "toDoc"; 
	/** xml-element for one sentence-level alignment
	 * <br/>examples for alignment types:
	 * <br/>1-1: &lt;link id="SL168" xtargets="169;175" /&gt;
	 * <br/>2-1: &lt;link id="SL169" xtargets="170 171;176" /&gt;
	 * <br/>1-2: &lt;link id="SL155" xtargets="156;161 162" /&gt;
	 * <br/>0-1: &lt;link id="SL170" xtargets=";177" /&gt;
	 * <br/>1-0: &lt;link id="SL614" xtargets="599;" /&gt;
	 * <br/>1-3: &lt;link id="SL623" xtargets="605;661 662 663" /&gt;
	 */
	static final String ELEMENT_link = "link"; 
	/** xml-attribute for the id of a sentence-level alignment */
	static final String ATTRIBUTE_id = "id";
	/** xml-attribute for the sentence-level alignment: 
	 * sentence number(s) of source and corresponding sentence number(s) of target
	 * separated by a semicolon */
	static final String ATTRIBUTE_xtargets = "xtargets"; 

	/** pattern for a line in a link group */
	static final Pattern PATTERN_linkGroup_toFrom = Pattern.compile("^<linkGrp targType=\"(.+)\" toDoc=\"(.+)\" fromDoc=\"(.+)\">$");
	/** pattern for a line in a link group */
	static final Pattern PATTERN_linkGroup_fromTo = Pattern.compile("^<linkGrp targType=\"(.+)\" fromDoc=\"(.+)\" toDoc=\"(.+)\">$");
	/** number of the specified group for the link-group pattern */
	static final int GROUP_linkGroup_toDocFirst = 2;
	/** number of the specified group for the link-group pattern */
	static final int GROUP_linkGroup_fromDocSecond = 3;
	/** number of the specified group for the link-group pattern */
	static final int GROUP_linkGroup_toDocSecond = 3;
	/** number of the specified group for the link-group pattern */
	static final int GROUP_linkGroup_fromDocFirst = 2;


	/* xml-element for sentence in the monolingual opus-files */
	static final String ELEMENT_s = "s"; //sentence
	/* xml-element for word in the monolingual opus-files */
	static final String ELEMENT_w = "w"; //word

	/** separator character between the sentence number(s) of the source text 
	 and the sentence number(s) of the target text in an alignment */
	public static final String SEPARATOR_opus_link = ";";

}
