/**
 * CRFalign: A CRF-based sentence aligner for parallel corpora.
 * 
 * Copyright (C) 2012  Eva Mujdricza-Maydt
 * Copyright (C) 2012  Huiqin Körkel-Qu
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Developers:
 * - Eva Mujdricza-Maydt (mujdricz@cl.uni-heidelberg.de)
 * - Huiqin Körkel-Qu (huiqin@cl.uni-heidelberg.de)
 */
 
package corpora.opensubtitles;

import com.cybozu.labs.langdetect.DetectorFactory;
import com.cybozu.labs.langdetect.LangDetectException;
import java.io.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.zip.GZIPInputStream;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import objects.Alignment_DocLevel;
import objects.Sentence;
import objects.Word;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/** Extracting OpenSubtitles (or OPUS) data 
 * (only opus-specified extraction, i.e. preprocessing).
 *
 * @author Eva Mujdricza-Maydt
 * @lastChange 20120626
 */
public class ExtractOpensubltitles {

	
	/** text from gzipped xml-file of one of the language-pair. 
	 * (e.g. de/1913/13044/223938_1of1.xml.gz)
	 */
	static ArrayList<Sentence> monolingualText; 
	/**
	 * entails possible clitica in opus texts. 
	 * <br/>NOTE that these clitica are not only "valid" clitica but all
	 * strange ones found in the texts
	 */
	private static ArrayList<String> clitics; 
	
	
	/** Copies all relevant OPUS files in a folder structure which can be 
	 * handled by the program.
	 * This method outputs the following folder structure: 
	 * <tt><br/>/path/to/dataSet/
	 * <br/>&nbsp;&nbsp;+--de/
	 * <br/>&nbsp;&nbsp;+--en/
	 * <br/>&nbsp;&nbsp;+--align/</tt>
	 * <br/>The original OPUS data has the following structure:
	 * <tt><br/>/path/to/opus/
	 * <br/>&nbsp;&nbsp;+--de-en/
	 * <br/>&nbsp;&nbsp;&nbsp;&nbsp;+--de/
	 * <br/>&nbsp;&nbsp;&nbsp;&nbsp;+--en/
	 * <br/>&nbsp;&nbsp;&nbsp;&nbsp;+--de-en.xml.gz</tt>
	 * 
	 * @param pathToOpus path up to the deepest shared folder for source and target
	 *					data in the original OPUS corpus. For the german-english
	 *					portion of data, this is the "de-en" folder.
	 * @param yearFrom the first year from the OPUS corpus you need
	 * @param yearTo  the last year from the OPUS corpus you need
	 * @param pathToDataSet output data set path
	 * @param libPath path to the lib folder
	 */
	public static void normalizeOpusCorpus(
		String pathToOpus, String yearFrom, String yearTo, String pathToDataSet,
		String libPath) 
		throws IOException, FileNotFoundException, ParserConfigurationException, 
		SAXException, LangDetectException {
	
		int iYearFrom = Integer.valueOf(yearFrom);
		int iYearTo = Integer.valueOf(yearTo);
		
		// /path/to/opus/de-en/
		//			     +- de-en.xml.gz
		//			     +- de/
		//			     +- en/
		pathToOpus = utils.Helper.normalizePath(pathToOpus);
		pathToDataSet = utils.Helper.normalizePath(pathToDataSet);
		libPath = utils.Helper.normalizePath(libPath);
		
		System.out.println("pathToOpus: " + pathToOpus);
		System.out.println("pathToDataSet: " + pathToDataSet);
		System.out.println("libPath: " + libPath);
		
		String alignFolder = main.Constants_Main.FOLDER_ALIGNMENTS;
		//- First, correct the gunzipped de-en.xml file (or other <src>-<trg>.xml), which entails the alignments between the file-pairs. The correction is needed, because the file is not a valid xml-file.
		//NOTE: The new xml-file is valid, it is not needed to correct them!
		String src = main.Constants_Main.LANG_SRC;
		String trg = main.Constants_Main.LANG_TRG;
		String alignmentsFileName_gz = src + "-" + trg + main.Constants_Main.EXT_XMLGZ;
		String alignmentsFileName = src + "-" + trg + main.Constants_Main.EXT_XML;
		if (!utils.Helper.hasFile(pathToOpus, alignmentsFileName)) {
			System.out.println("ungzip alignmentsFile: " + pathToOpus + alignmentsFileName_gz);
			
			unzip_gz(pathToOpus + alignmentsFileName_gz, main.Constants_Main.EXT_GZ);
		}
		
		System.out.println("alignmentsFileName: " + alignmentsFileName);
		
		String originalAlignmentFileName = pathToOpus + alignmentsFileName;
		if (!(new File(originalAlignmentFileName).canRead())) {
			String msg = "\nAlignment file \"" + originalAlignmentFileName + "\" cannot be read.\n";
			throw new NullPointerException(msg);
		}
		//This part is not needed, because the new alignment file is valid.
//		String originalAlignmentFileName_wrong = originalAlignmentFileName + "_wrong";
//		if ((!new File(originalAlignmentFileName_wrong).canRead())) {
//			System.out.println(">correct " + alignmentsFileName + " file");
//			utils.Helper.copyFile(originalAlignmentFileName, originalAlignmentFileName_wrong);
//			correctOriginalOpusAlignmentFile(originalAlignmentFileName_wrong, originalAlignmentFileName);
//		}
		
		//- Unzip the files to xml.
		System.out.println(">unzipMovieFiles");
		unzipMovieFiles(pathToOpus, pathToOpus, alignmentsFileName);
		
		System.out.println(">writePlainTexts_tokens_withCorrectedClitics");
		//- Extract the plain text file for all files from the xml-files: with extension .text_tokens.txt
		//(this is needed, because we want to proof, whether the language of the file is appropriate, and the language detector takes a plain text file as input). Also the xml-residue-detector will be applied to the plain text.
		writePlainTexts_tokens_withCorrectedClitics(pathToOpus, pathToOpus, alignmentsFileName);
		
		System.out.println(">getSrcToTrgMovieTitlesFromAlignAnnotation_withFiltering");

		HashMap<String, String> filteredSrcToTrgFileNames = 
			getSrcToTrgMovieTitlesFromAlignAnnotation_withFiltering(
			pathToOpus, libPath, pathToOpus, alignmentsFileName);
		System.out.println("--> got " + filteredSrcToTrgFileNames.size() + " file-pairs.");
		
		//make src and trg subfolders if needed
		System.out.println(">createSubfolderIfNeeded");
		pathToDataSet = (new File(pathToDataSet)).getCanonicalPath();
		String pathTo = main.Constants_Main.SLASH;
		for (String p : pathToDataSet.split(main.Constants_Main.SLASH)) {
			if (p.isEmpty()) {
				continue;
			}
			utils.Helper.createSubfolderIfNeeded(pathTo, p);
			pathTo = pathTo + p + main.Constants_Main.SLASH;
		}
		pathToDataSet = utils.Helper.normalizePath(pathToDataSet);
		utils.Helper.createSubfolderIfNeeded(pathToDataSet, src);
		utils.Helper.createSubfolderIfNeeded(pathToDataSet, trg);
		utils.Helper.createSubfolderIfNeeded(pathToDataSet, main.Constants_Main.ALIGNMENTS);
		
		for (String srcFileName : filteredSrcToTrgFileNames.keySet()) {
			
			if (!(getYear(srcFileName) >= iYearFrom && getYear(srcFileName) <= iYearTo)) {
				continue;
			}
			
			String trgFileName = filteredSrcToTrgFileNames.get(srcFileName);
			
			String srcFileStem = utils.Helper.changeFileNameExtension(srcFileName, main.Constants_Main.EXT_XMLGZ, "");
			String trgFileStem = utils.Helper.changeFileNameExtension(trgFileName, main.Constants_Main.EXT_XMLGZ, "");
			
			String srcFilePartStem = srcFileStem.substring(srcFileStem.lastIndexOf(main.Constants_Main.SLASH)+1);
			String trgFilePartStem = trgFileStem.substring(trgFileStem.lastIndexOf(main.Constants_Main.SLASH)+1);
			
			String srcTrgFileName = srcFilePartStem + main.Constants_Main.SEPARATOR_SRCTRG_FILENAME + trgFilePartStem + main.Constants_Main.EXT_TXT;
			
			//copy src-file	
			System.out.println("SRC COPY FROM: " + pathToOpus + srcFileStem + main.Constants_Main.EXT_TEXT_TOKENS);
			System.out.println("SRC COPY TO: " + pathToDataSet + src + main.Constants_Main.SLASH + srcTrgFileName);
			utils.Helper.copyFile(pathToOpus + srcFileStem + main.Constants_Main.EXT_TEXT_TOKENS,
				pathToDataSet + src + main.Constants_Main.SLASH + srcTrgFileName);
			//copy trg-file
			System.out.println("TRG COPY FROM: " + pathToOpus + trgFileStem + main.Constants_Main.EXT_TEXT_TOKENS);
			System.out.println("TRG COPY TO: " + pathToDataSet + trg + main.Constants_Main.SLASH + srcTrgFileName);
			
			utils.Helper.copyFile(pathToOpus + trgFileStem + main.Constants_Main.EXT_TEXT_TOKENS,
				pathToDataSet + trg + main.Constants_Main.SLASH + srcTrgFileName);
			//write align-file
			String opusAlignmentsOutputFileName = utils.Helper.changeFileNameExtension(
				srcTrgFileName, main.Constants_Main.EXT_TXT, main.Constants_Main.EXT_OPUSALIGNMENT);
			String opusAlignmentsOutputFileNameWithPath = pathToDataSet + alignFolder
				 + opusAlignmentsOutputFileName;
			Alignment_DocLevel currAlignments = extractOpusAlignment_xml(
				originalAlignmentFileName, srcFileName);
			corpora.ExtractCorpus_forEvaluation.writeAlignmentFile_opusFormat(
				opusAlignmentsOutputFileNameWithPath, currAlignments);
		}
	}

	/** Writes the tokens of the texts with corrected clitica.
	 * <br/>NOTE that after this point, the files generated by this method should be 
	 * taken for reading the text.
	 * 
	 * @param opusPath path to the OPUS corpus
	 * @param alignPath path to the OPUS alignment xml-file
	 * @param alignAnnotFileName file name of the OPUS alignment xml-file (de-en.xml)
	 * @throws FileNotFoundException
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException 
	 */
	private static void writePlainTexts_tokens_withCorrectedClitics(
		String opusPath, String alignPath, String alignAnnotFileName) 
		throws FileNotFoundException, IOException, 
		ParserConfigurationException, SAXException {
		
		System.out.println("corpusPath: " + opusPath);
		System.out.println("alignPath: " + alignPath);
		System.out.println("alignAnnotFileName: " + alignAnnotFileName);
		
		HashMap<String, String> srcToTrgMovies = getSrcToTrgMovieTitlesFromAlignAnnotation(
			opusPath, alignPath, alignAnnotFileName);
		System.out.println("There are " + srcToTrgMovies.size() + " detected movie pairs.");

		for (String srcFullName : utils.Helper.sortKeys_str(srcToTrgMovies.keySet())) {
			
			String src = utils.Helper.changeFileNameExtension(srcFullName, main.Constants_Main.EXT_GZ, "");
			ArrayList<Sentence> textInSentences_src = extractMonolingualText_opus_xml(opusPath + src);
			textInSentences_src = correctClitics(textInSentences_src);
			String plainTextFileName_src = opusPath + utils.Helper.changeFileNameExtension(src, main.Constants_Main.EXT_XML, main.Constants_Main.EXT_TEXT_TOKENS);
			corpora.ExtractCorpus.writePlainText_tokens(textInSentences_src, plainTextFileName_src);

			String trg = utils.Helper.changeFileNameExtension(srcToTrgMovies.get(srcFullName), main.Constants_Main.EXT_GZ, "");
			ArrayList<Sentence> textInSentences_trg = extractMonolingualText_opus_xml(opusPath + trg);
			textInSentences_trg = correctClitics(textInSentences_trg);
			String plainTextFileName_trg = opusPath + utils.Helper.changeFileNameExtension(trg, main.Constants_Main.EXT_XML, main.Constants_Main.EXT_TEXT_TOKENS);
			corpora.ExtractCorpus.writePlainText_tokens(textInSentences_trg, plainTextFileName_trg);
			System.out.print(".");
			
		}
		System.out.println();
	}
	
	
	/** Unzips the original gzipped movie files.
	 * 
	 * @param opusPath path to the OPUS corpus
	 * @param alignPath path to the OPUS alignment xml-file
	 * @param alignAnnotFileName file name of the OPUS alignment xml-file (de-en.xml)
	 * @throws FileNotFoundException
	 * @throws IOException 
	 */
	private static void unzipMovieFiles(
		String opusPath, 
		String alignPath, 
		String alignAnnotFileName) 
			throws FileNotFoundException, IOException {
		
		System.out.println("corpusPath: " + opusPath);
		System.out.println("alignPath: " + alignPath);
		System.out.println("alignAnnotFileName: " + alignAnnotFileName);
		
		HashMap<String, String> srcToTrgFiles = getSrcToTrgMovieTitlesFromAlignAnnotation(
				opusPath, alignPath, alignAnnotFileName);

		System.out.println("SOURCE FILES: " + srcToTrgFiles.size());
		for (String s : utils.Helper.sortKeys_str(srcToTrgFiles.keySet())) {
			String inputFileName = opusPath + s;
			String xmlFile = unzip_gz(inputFileName, main.Constants_Main.EXT_GZ);
			System.out.print(".");
		}
		System.out.println();
		System.out.println("TARGET FILES: " + srcToTrgFiles.size());
		for (String s : utils.Helper.sortValues_str(srcToTrgFiles.values())) {
			String inputFileName = opusPath + s;
			String xmlFile = unzip_gz(inputFileName, main.Constants_Main.EXT_GZ);
			System.out.print(".");
		}
		System.out.println();
	}

	/** Extracts opus text from the original, monolingual xml-files into a
	 * <code>Sentence</code> list.
	 * 
	 * @param inputFile xml-file with the text of the movie subtitles
	 * @return	list of <code>Sentence</code>s with the text
	 * @throws FileNotFoundException
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException 
	 */
	private static ArrayList<Sentence> extractMonolingualText_opus_xml(
		String inputFile)
		throws FileNotFoundException, IOException,
		ParserConfigurationException, SAXException {

		try {
			SAXParserFactory factory = SAXParserFactory.newInstance();
			SAXParser saxParser = factory.newSAXParser();
			DefaultHandler handler = new XMLHandler_opensubtitles_monolingualXML();

			try {
				File f = new File(inputFile);
				saxParser.parse(f, handler);

			} catch (IOException ioex) {
				System.err.println("File " + inputFile + " not found.");
				ioex.printStackTrace();
				System.exit(2);
			}

		} catch (ParserConfigurationException pcex) {
			pcex.printStackTrace();
			System.exit(2);
		} catch (SAXException saxex) {
			saxex.printStackTrace();
			System.exit(2);
		}

		return monolingualText;
	}

//	/** The original alignment file with the opus-alignments will be rewrite
//	 * to be a valid xml-file.
//	 * 
//	 * @param alignmentFileName	the name of the original alignment file
//	 *				&lt;SOURCELANG&gt;-&lt;TARGETLANG&gt;.xml
//	 */
//	private static void correctOriginalOpusAlignmentFile(
//		String alignmentFileName, String correctedAlignmentFileName)
//		throws FileNotFoundException, IOException {
//
//		//check
//		utils.Helper.checkFileForReading(alignmentFileName);
//		utils.Helper.checkFileForWriting(correctedAlignmentFileName);
//
//		BufferedReader br = new BufferedReader(new FileReader(new File(alignmentFileName)));
//		FileWriter fw = new FileWriter(new File(correctedAlignmentFileName));
//
//		boolean first = true;
//		String endTag = "</" + Constants_Opus.ELEMENT_linkGrp + ">";
//		String line = null;
//		while((line = br.readLine()) != null) {
//			if (line.startsWith("<" + Constants_Opus.ELEMENT_linkGrp)) {
//				if (first == true) {
//					fw.write(line + "\n");
//					first = false;
//				} else {
//					//end the previous linkGrp
//					line = endTag + "\n" + line;
//					fw.write(line + "\n");
//				}
//				
//			} else if (line.trim().equals("</cesAlign>")) {
//				line = endTag + "\n" + line;
//				fw.write(line + "\n");
//			} else if (line.startsWith("<!DOCTYPE")) {
//				//no write
//			} else {
//				fw.write(line + "\n");
//			}
//			
//		}
//		br.close();
//		fw.close();
//	}
	


	/** extracts aligmnents for one document pair
	 * <p>NOTE that the given DTD will be ignored</p>
	 * 
	 * @param alignmentFileName original Opus xml-file with alignments for all 
	 *							provided source-target files
	 * @param srcFileName file name of the current source file
	 * @return	all alignments for a source-target file pair
	 * @throws FileNotFoundException
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException 
	 */
	private static Alignment_DocLevel extractOpusAlignment_xml(
		String alignmentFileName, String srcFileName)
		throws FileNotFoundException, IOException,
		ParserConfigurationException, SAXException {

		utils.Helper.checkFileForReading(alignmentFileName);
		
		corpora.ExtractCorpus.srcFileNameForAligning = srcFileName;
		try {
			SAXParserFactory factory = SAXParserFactory.newInstance();
			SAXParser saxParser = factory.newSAXParser();
			//see http://www.isocra.com/2006/05/making-xerces-ignore-a-dtd/
			final org.xml.sax.XMLReader parser = saxParser.getXMLReader();
			parser.setFeature("http://apache.org/xml/features/nonvalidating/load-external-dtd", false);
			parser.setFeature("http://xml.org/sax/features/validation", false);

			DefaultHandler handler = new XMLHandler_opensubtitles_alignmentsXML_oneDocPair();
			
			parser.setContentHandler(handler);
			parser.parse(alignmentFileName);

		} catch (ParserConfigurationException pcex) {
			pcex.printStackTrace();
			System.exit(2);
		} catch (SAXException saxex) {
			saxex.printStackTrace();
			System.exit(2);
		}

		return corpora.ExtractCorpus.corpusAlignment_oneDoc;
	}

	
	/** Gets the map of source and corresponding target movie file names.
	 * 
	 * @param corpusPath path to the OPUS corpus
	 * @param alignPath path to the OPUS alignment xml-file
	 * @param alignAnnotFileName file name of the OPUS alignment xml-file (de-en.xml)
	 * @return map of source and corresponding target movie file names as they 
	 *			are written in the OPUS alignment xml-file
	 * @throws FileNotFoundException
	 * @throws IOException 
	 */
	private static HashMap<String, String> getSrcToTrgMovieTitlesFromAlignAnnotation(
		String corpusPath, String alignPath, String alignAnnotFileName)
		throws FileNotFoundException, IOException {

		HashMap<String, String> srcToTrgMap = new HashMap<String, String>();

		//check
		utils.Helper.checkFileForReading(alignPath + alignAnnotFileName);
		BufferedReader br = new BufferedReader(new FileReader(new File(alignPath + alignAnnotFileName)));
		String line = null;
		while ((line = br.readLine()) != null) {

			//example: <linkGrp targType="s"
			//toDoc="en/Action/1984/3246_62585_80503_beverly_hills_cop.xml.gz"
			//fromDoc="de/Action/1984/3246_62587_80507_beverly_hills_cop.xml.gz">
			if (line.startsWith("<" + Constants_OpenSubtitles.ELEMENT_linkGrp)) {
				Matcher mToFrom = Constants_OpenSubtitles.PATTERN_linkGroup_toFrom.matcher(line);
				Matcher mFromTo = Constants_OpenSubtitles.PATTERN_linkGroup_fromTo.matcher(line);
				if (mToFrom.matches() || mFromTo.matches()) {
					String fromDoc = null;
					String toDoc = null;
					if (mToFrom.matches()) {
						fromDoc = mToFrom.group(Constants_OpenSubtitles.GROUP_linkGroup_fromDocSecond);
						toDoc = mToFrom.group(Constants_OpenSubtitles.GROUP_linkGroup_toDocFirst);
					} else {
						fromDoc = mFromTo.group(Constants_OpenSubtitles.GROUP_linkGroup_fromDocFirst);
						toDoc = mFromTo.group(Constants_OpenSubtitles.GROUP_linkGroup_toDocSecond);
					}

					boolean exist = true;
					try {
						utils.Helper.checkFileForReading(corpusPath + fromDoc);
						utils.Helper.checkFileForReading(corpusPath + toDoc);
					} catch (FileNotFoundException fne) {
						exist = false;
					} catch (IOException ioe) {
						exist = false;
					}

					if (exist) {
						srcToTrgMap.put(fromDoc, toDoc);
					}
				} else {
					String msg = "\nNo match for line:\n" + line + "\n";
					throw new NullPointerException(msg);
				}
			}
		}
		br.close();

		return srcToTrgMap;
	}
	
	/** extracts the appropriate parallel files from the original alignment annotation.
	 * There are files, which are existing, have a corresponding file pare,
	 * are written in the given language, and do not entail xml-residues.
	 * 
	 * @param corpusPath path to opus corpus
	 * @param libPath path to the lib-folder with the profiles
	 * @param alignPath path to the alignAnnotFileName (within the corpusPath)
	 * @param alignAnnotFileName for opus corpus <src>-<trg>.xml
	 * @return map of appropriate source and corresponding target movie file names as they 
	 *			are written in the OPUS alignment xml-file
	 * @throws FileNotFoundException
	 * @throws IOException
	 * @throws LangDetectException 
	 */
	private static HashMap<String, String> getSrcToTrgMovieTitlesFromAlignAnnotation_withFiltering(
		String corpusPath, String libPath, String alignPath, String alignAnnotFileName)
		throws FileNotFoundException, IOException, LangDetectException {

		//for language filtering
		String pathProfile = libPath + annotators.languageDetector.Constants_LanguageDetector.FOLDER_PROFILE;
		System.out.println("pathProfile: " + pathProfile);
		
		DetectorFactory.loadProfile(pathProfile);

		HashMap<String, String> srcToTrgMap = new HashMap<String, String>();

		//check
		utils.Helper.checkFileForReading(alignPath + alignAnnotFileName);
		BufferedReader br = new BufferedReader(new FileReader(new File(alignPath + alignAnnotFileName)));
		String line = null;
		while ((line = br.readLine()) != null) {

			//example: <linkGrp targType="s"
			//toDoc="en/Action/1984/3246_62585_80503_beverly_hills_cop.xml.gz"
			//fromDoc="de/Action/1984/3246_62587_80507_beverly_hills_cop.xml.gz">
			if (line.startsWith("<" + Constants_OpenSubtitles.ELEMENT_linkGrp)) {
				Matcher mToFrom = Constants_OpenSubtitles.PATTERN_linkGroup_toFrom.matcher(line);
				Matcher mFromTo = Constants_OpenSubtitles.PATTERN_linkGroup_fromTo.matcher(line);
				if (mToFrom.matches() || mFromTo.matches()) {
					String fromDoc = null;
					String toDoc = null;
					if (mToFrom.matches()) {
						fromDoc = mToFrom.group(Constants_OpenSubtitles.GROUP_linkGroup_fromDocSecond);
						toDoc = mToFrom.group(Constants_OpenSubtitles.GROUP_linkGroup_toDocFirst);
					} else {
						fromDoc = mFromTo.group(Constants_OpenSubtitles.GROUP_linkGroup_fromDocFirst);
						toDoc = mFromTo.group(Constants_OpenSubtitles.GROUP_linkGroup_toDocSecond);
					}

					boolean exist = true;
					try {
						utils.Helper.checkFileForReading(corpusPath + fromDoc);
						utils.Helper.checkFileForReading(corpusPath + toDoc);
					} catch (FileNotFoundException fne) {
						exist = false;
					} catch (IOException ioe) {
						exist = false;
					}
					//NOTE that not all files exist, for which there is alignment information,
					//so we have to ignore alignment information for these non-existing files
					if (!exist) {
						continue;
					}

					String src_text = utils.Helper.changeFileNameExtension(fromDoc, main.Constants_Main.EXT_XMLGZ, main.Constants_Main.EXT_TEXT_TOKENS);
					String trg_text = utils.Helper.changeFileNameExtension(toDoc, main.Constants_Main.EXT_XMLGZ, main.Constants_Main.EXT_TEXT_TOKENS);
					
					String srcLang = annotators.languageDetector.DetectLanguage.detectLanguageOfDocument(corpusPath + src_text, pathProfile);
					String trgLang = annotators.languageDetector.DetectLanguage.detectLanguageOfDocument(corpusPath + trg_text, pathProfile);
					//files with not the given language will be filtered out
					if (!(srcLang.equals(main.Constants_Main.LANG_SRC) && trgLang.equals(main.Constants_Main.LANG_TRG))) {
						continue;
					}

					//files with XML-tags will be filtered out
					if (hasXMLResidue(corpusPath + src_text) || hasXMLResidue(corpusPath + trg_text)) {
						continue;
					}

					srcToTrgMap.put(fromDoc, toDoc);

				} else {
					String msg = "\nNo match for line:\n" + line + "\n";
					throw new NullPointerException(msg);
				}
			}
		}
		br.close();

		return srcToTrgMap;
	}
	
	

	/** Checks whether a file entails html or other xml annotation.
	 *<br/>NOTE that this method returns true even if there is only one &lt; at the 
	 * beginning of a line. 
	 * 
	 * @param inputFileName text file name with the plain text 
	 * @return	{@code true} if the text entails residue of xml tags, 
	 *			{@code false} otherwise
	 * @throws FileNotFoundException
	 * @throws IOException 
	 */
	private static boolean hasXMLResidue(
		String inputFileName) 
		throws FileNotFoundException, IOException {
		
		int threshold = 0;	//if 0, no xml-residues will be allowed
		int count = 0;
		
		BufferedReader br = new BufferedReader(new FileReader(new File(inputFileName)));
		String line = null;
		while ((line = br.readLine()) != null) {
			
			if (((line.indexOf("<") >= 0) || (line.indexOf(">") >= 0))
				&& !((line.indexOf("<<") >= 0) || (line.indexOf(">>") >= 0))) {
				
				count++;
				if (count > threshold) {
					break;
				}
			}
		}
		br.close();
		
		return count > threshold;
	}
	
	/** The original text files for OpenSubtitle (OPUS) data entail many incorrect
	 * clitic splitting. E.g. "I' m" instead of "I 'm". These mistakes will be
	 * here corrected in order that TreeTagger can tag the text correctly.
	 * 
	 * @param sentences	list of sentences in the text
	 * @return	list of sentences with corrected clitica
	 */
	private static ArrayList<Sentence> correctClitics(ArrayList<Sentence> sentences) {

		for (Sentence sent : sentences) {
			ArrayList<Word> words = sent.getWordObjList_suppressException();
			for (int i = 0; i < words.size() - 1; i++) { //NOTE that the loop is only till the one but last element
				String w = words.get(i).getWord();
				if ((w.length() > 1) && w.endsWith("'") && isClitic(words.get(i + 1).getWord())) {
					words.get(i).setWord(w.substring(0, w.length() - 1));
					String cl = words.get(i + 1).getWord();
					words.get(i + 1).setWord("'" + cl);
				}
			}
		}

		return sentences;
	}

	/** Checks whether a token is a clitic word. 
	 * 
	 * @param token given token
	 * @return {@code true} if the token is found in a clitic list, 
	 *			{@code false} otherwise
	 */
	private static boolean isClitic(String token) {
		boolean is = false;
		clitics = loadClitics();
		if (clitics.contains(token)) {
			is = true;
		}
		return is;
	}

	/** The list of clitics are heuristically determined.
	 * 
	 * @return list of clitics in OpenSubtitles (OPUS) corpus
	 */
	private static ArrayList<String> loadClitics() {
		if (clitics == null) {
			clitics = new ArrayList<String>();
			clitics.add("m");
			clitics.add("re");
			clitics.add("s");
			clitics.add("il");
			clitics.add("ill");
			clitics.add("ll");
			clitics.add("ve");
			clitics.add("d");
			clitics.add("t");
		}
		return clitics;
	}

	
	
	/** Get the year information from the given file path.
	 * e.g. If the file name given by OPUS is "de/2009/19984/3679357_1of1.xml.gz"
	 * then we want to take "2009".
	 * 
	 * @param opusTitleName name of the file with a relative path given by OPUS
	 *						alignment xml-file
	 * @return the year information extracted from the file name
	 */
	private static int getYear(String opusTitleName) {
		
		int i1 = opusTitleName.indexOf(main.Constants_Main.SLASH);
		if (i1 == -1) {
			String msg = "\nNo slash found in the movie title name \"" + opusTitleName + "\"\n";
			throw new NullPointerException(msg);
		}
		int i2 = opusTitleName.indexOf(main.Constants_Main.SLASH, i1+1);
		if (i2 == -1) {
			String msg = "\nNo second slash found in the movie title name \"" + opusTitleName + "\"\n";
			throw new NullPointerException(msg);
		}
		
		int year = Integer.valueOf(opusTitleName.substring(i1+1, i2));
		return year;
	}
	
	/** Unzip files compressed to gz.
	 * 
	 * @param gzippedFileName compressed input file name
	 * @param zipEnding the compressing extension (".gz")
	 * @return the decompressed file without the zipEnding in its file name
	 * @throws FileNotFoundException
	 * @throws IOException 
	 */
	private static String unzip_gz(String gzippedFileName, String zipEnding)
		throws FileNotFoundException, IOException {

		String unzippedFileName = null;
		if (!gzippedFileName.endsWith(zipEnding)) {
			String msg = "\nThe given filename " + gzippedFileName + " does not ends with " + zipEnding + "\n";
			throw new NullPointerException(msg);
		} else {
			unzippedFileName = gzippedFileName.substring(0, (gzippedFileName.length()-zipEnding.length()));
		}

		InputStream  in = null;
		OutputStream out = null;

		try {
			in = new GZIPInputStream(new FileInputStream(gzippedFileName));
			out  = new FileOutputStream(unzippedFileName);

			byte[] buffer = new byte[1024];
			int len;
			while ((len = in.read(buffer)) > 0) {
				out.write(buffer, 0, len);
			}
			in.close();
			out.close();
		}
		catch (IOException e) {
			System.err.println("File " + gzippedFileName + " cannot be ungziped.\n");
			e.printStackTrace();
			System.exit(2);
		}
		//check
		utils.Helper.checkFileForReading(unzippedFileName);

		return unzippedFileName;
	}
	
}

