/**
 * CRFalign: A CRF-based sentence aligner for parallel corpora.
 * 
 * Copyright (C) 2012  Eva Mujdricza-Maydt
 * Copyright (C) 2012  Huiqin Körkel-Qu
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Developers:
 * - Eva Mujdricza-Maydt (mujdricz@cl.uni-heidelberg.de)
 * - Huiqin Körkel-Qu (huiqin@cl.uni-heidelberg.de)
 */
package objects;

import java.util.ArrayList;

/** This object bond together all sentence-level alignments 
 * for a source-target document pair.
 * <p/>
 * NOTE that "sentence-level alignments" mean that sentences will be aligned.
 * One sentence-level alignment entails one or more sentence number(s) 
 * of the source document and one or more sentence number(s) of the target 
 * document. The sentences with the given numbers are aligned in the given 
 * alignment. If a source sentence has no counterpart in the target document,
 * then on the target side, an empty list will be stored. If a target sentence has
 * no counterpart in the source document, then the source sentence number list
 * keeps empty. These two cases are so called "zero alignments". NOTE that 
 * in a zero alignment only one sentence will be bounded to "zero", even if there
 * are more sentences one after another which are aligned with no other sentences.
 *
 * @author Eva Mujdricza-Maydt
 * @lastChange 20120720
 */
public final class Alignment_DocLevel {

	private String document1; //source document name
	private String document2; //target document name

	private ArrayList<Alignment_SentLevel> alignments; //list of sentence alignments

	
	/** Constructor given the source and target document name.
	 * 
	 * @param srcDoc name of the source document
	 * @param trgDoc name of the target document
	 */
	public Alignment_DocLevel(String srcDoc, String trgDoc) {
		this.document1 = srcDoc;
		this.document2 = trgDoc;
	}

///////////////////////////////////////////////////////////////////////////////

	/** Checks whether the source document name is set.
	 * 
	 * @return {@code true} if the source document name is a non-empty 
	 *			{@code String}, {@code false} otherwise.
	 */
	private boolean hasSourceDocument() {
		boolean has = true;
		if ((this.document1 == null) || this.document1.isEmpty()) {
			has = false;
		}
		return has;
	}
	
	/**
	 * Gets the name of the source document.
	 * 
	 * @return the name of the source document will be returned, or a 
	 *			{@code NullPointerException} will be thrown if the name of the 
	 *			source document is not set.
	 */
	public String getSourceDocument() {
		if (!hasSourceDocument()) {
			String msg = "\nNo source document name defined.\n";
			throw new NullPointerException(msg);
		}
		return this.document1;
	}

	/** Checks whether the target document name is set.
	 * 
	 * @return {@code true} if the target document name is a non-empty 
	 *			String, {@code false} otherwise.
	 */
	private boolean hasTargetDocument() {
		boolean has = true;
		if ((this.document2 == null) || this.document2.isEmpty()) {
			has = false;
		}
		return has;
	}

	/**
	 * Gets the name of the target document.
	 * 
	 * @return the name of the target document will be returned, or a 
	 *			{@code NullPointerException} will be thrown if the name of the 
	 *			target document is not set.
	 */
	public String getTargetDocument() {
		if (!hasTargetDocument()) {
			String msg = "\nNo target document name defined.\n";
			throw new NullPointerException(msg);
		}
		return this.document2;
	}


	/** Sets a list of sentence-level alignments.
	 * 
	 * @param sa 
	 */
	public void setSentAlignments(ArrayList<Alignment_SentLevel> sa) {
		this.alignments = sa;
	}

	/** Adds a new sentence-level alignment to the alignments for the 
	 * given document pair.
	 * 
	 * @param newAl	additional sentence-level alignment for the given 
	 *				document pair
	 */
	public void addSentAlignment(Alignment_SentLevel newAl) {
		if (this.alignments == null) {
			this.alignments = new ArrayList<Alignment_SentLevel>();
		}
		this.alignments.add(newAl);
	}

	/** Checks whether the alignments are initialized.
	 * 
	 * @return {@code true} if the list of sentence-level alignments are 
	 *			initialized, {@code false} otherwise.
	 */
	private boolean hasSentAlignments() {
		boolean has = true;
		if (this.alignments == null) {
			has = false;
		}
		return has;
	}

	/** Gets the list of sentence-level alignments for the given document
	 * pair.
	 * 
	 * @return the list of sentence-level alignments; if no alignments available,
	 *			a {@code NullPointerException} will be thrown
	 */
	public ArrayList<Alignment_SentLevel> getSentAlignments() {
		if (!hasSentAlignments()) {
			String msg = "\nNo alignment for documents \"" + getSourceDocument() + "\" and \"" + getTargetDocument() + "\"\n";
			throw new NullPointerException(msg);
		}

		return this.alignments;
	}

}
