/**
 * CRFalign: A CRF-based sentence aligner for parallel corpora.
 * 
 * Copyright (C) 2012  Eva Mujdricza-Maydt
 * Copyright (C) 2012  Huiqin Körkel-Qu
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Developers:
 * - Eva Mujdricza-Maydt (mujdricz@cl.uni-heidelberg.de)
 * - Huiqin Körkel-Qu (huiqin@cl.uni-heidelberg.de)
 */
package objects;

import java.util.ArrayList;

/** Representation of a sentence in a text.
 * In the corpora we use, one line entails one sentence. 
 * <br/>NOTE that in OPUS data also empty lines are considered as sentences!
 *
 * @author Eva Mujdricza-Maydt
 */
public final class Sentence {

	/** number of the current sentence in the text. 
	 * The first sentence has the number 1. */
	private int sentNr;
	/** list of {@code Word} objects in the current sentence. */
	private ArrayList<Word> wordList;

	/** Constructor */
	public Sentence() {
		sentNr = 0;
		wordList = null;
	}


///////////////////////////////////////////////////////////////////////////////

	/** Sets the number of the sentence in the text.
	 * 
	 * @param sn sentence number (lowest allowed sentence number: 1)
	 */
	public void setSentenceNr(int sn) {
		sentNr = sn;
	}

	/** Checks whether the number of the sentence is set.
	 * 
	 * @return {@code true} if the sentence number is set (that is it is a 
	 *		number higher than zero), {@code false} otherwise
	 */
	private boolean hasSentenceNr() {
		boolean has = true;
		if (this.sentNr == 0) {
			has = false;
		}
		return has;
	}

	/** Get the number of the current sentence. 
	 * 
	 * @return the number of the sentence in the text
	 * @throws NullPointerException if no sentence number is set
	 */
	public int getSentenceNr() {
		if (hasSentenceNr() == false) {
			String msg = "\nNo sentence number defined.\n";
			throw new NullPointerException(msg);
		}
		return this.sentNr;
	}

	/** Sets the {@code Word} list for the current sentence.
	 * 
	 * @param wl list of {@code Word} objects containing the current sentence
	 */
	public void setWordList(ArrayList<Word> wl) {
		this.wordList = wl;
	}

	/** Checks whether the word list of the sentence is set.
	 * 
	 * @return {@code true} if the word list of the sentence is set, 
	 *			{@code false} otherwise
	 */
	private boolean hasWordList() {
		boolean has = true;
		if ((this.wordList == null)
			|| this.wordList.isEmpty()) {
			has = false;
		}
		return has;
	}

	/** Gets the list of {@code Word} objects containing the current sentence.
	 * 
	 * @return list of {@code Word} objects in the current sentence.
	 * @throws NullPointerException if the sentence contains no words at all 
	 *		(this is a possible case) 
	 */
	public ArrayList<Word> getWordObjList() {
		if (hasWordList() == false) {
			String msg = "\nNo word list defined in sentence " + getSentenceNr() + ".\n";
			throw new NullPointerException(msg);
		}
		return this.wordList;
	}
	
	/** Gets the list of {@code Word} objects containing the current sentence.
	 * This method also gets sentences with zero length, i.e. if the 
	 * sentence entails no words at all.
	 * 
	 * @return list of {@code Word} objects in the current sentence.
	 */
	public ArrayList<Word> getWordObjList_suppressException() {
		if (hasWordList() == false) {
			return new ArrayList<Word>();
		}
		return this.wordList;
	}

	
	/** Gets the current sentence as a String, containing its tokens.
	 * This method also gets sentences with zero length, i.e. if the 
	 * sentence entails no words at all.
	 * 
	 * @return a string with the tokens in the sentence
	 */
	public String getWordListAsString_suppressException() {
		
		StringBuilder sb = new StringBuilder();
		for (Word w : this.wordList) {
			sb.append(w.getWord()).append(" ");
		}
		return sb.toString().trim();
	}
	
}
